"""Unittest for IDAT preprocessing."""

import unittest

import numpy as np
import numpy.testing as npt

from mepylome.dtypes import ArrayType, Manifest, MethylData, RawData
from mepylome.tests.helpers import TempIdatFilePair, TempManifest


class TestIdatPreprocessing(unittest.TestCase):
    """Unittest for IDAT preprocessing."""

    def test_raw_data(self):
        tmp_manifest = TempManifest()
        manifest = Manifest(raw_path=tmp_manifest.path)
        manifest.array_type = ArrayType.UNKNOWN
        self._test_raw_data(54, manifest, 2)
        # self._test_raw_data(622500, Manifest("450k"), 1)
        # self._test_raw_data(622500, Manifest("450k"), 4)
        # self._test_raw_data(1051000, Manifest("epic"), 3)
        # self._test_raw_data(1104000, Manifest("epicv2"), 2)

        # Clean up
        manifest.proc_path.unlink()
        manifest.ctrl_path.unlink()
        manifest._pickle_path.unlink()

    def _test_raw_data(self, n_cpgs, manifest, n_probes):
        id_list = sorted(
            (
                set(manifest.data_frame.AddressA_ID)
                | set(manifest.data_frame.AddressB_ID)
                | set(manifest.control_data_frame.Address_ID)
            )
            - {-1}
        )
        id_list = id_list[:n_cpgs]
        ids = np.full(n_cpgs, -1, dtype="<i4")
        ids[: len(id_list)] = np.array(id_list, dtype="<i4")
        start_value = id_list[-1] + 1
        ids[len(id_list) :] = np.arange(
            start_value,
            start_value + (n_cpgs - len(id_list)),
            dtype="<i4",
        )

        # Create and test two sets of data
        idat_pairs = []
        for index in range(n_probes):
            mean_grn = ((ids + index) % 23).astype(dtype="<u2")
            mean_red = ((ids + index) % 21).astype(dtype="<u2")
            test_grn = {
                "n_snps_read": n_cpgs,
                "illumina_ids": ids,
                "probe_means": mean_grn,
            }
            test_red = {
                "n_snps_read": n_cpgs,
                "illumina_ids": ids,
                "probe_means": mean_red,
            }
            idat_pairs.append(
                TempIdatFilePair(data_grn=test_grn, data_red=test_red)
            )

        if manifest.array_type == ArrayType.UNKNOWN:
            raw_data = RawData(
                [file.basepath for file in idat_pairs], manifest=manifest
            )
        else:
            raw_data = RawData([file.basepath for file in idat_pairs])
        self.assertIsNotNone(str(raw_data))
        self.assertIsNotNone(str(raw_data.manifest))
        npt.assert_array_equal(
            raw_data.ids, ids, err_msg="Mismatch in IDs array"
        )
        npt.assert_array_equal(
            raw_data._grn,
            np.array([pair.data_grn["probe_means"] for pair in idat_pairs]),
            err_msg="Mismatch in Green channel array",
        )
        npt.assert_array_equal(
            raw_data._red,
            np.array([pair.data_red["probe_means"] for pair in idat_pairs]),
            err_msg="Mismatch in Red channel array",
        )
        self.assertEqual(
            raw_data.probes,
            [pair.basepath.name for pair in idat_pairs],
            "Mismatch in probe name",
        )
        self.assertEqual(
            raw_data.array_type, manifest.array_type, "Mismatch in array type"
        )

        if manifest.array_type == ArrayType.UNKNOWN:
            self._test_methyl_data_raw(raw_data)
            self._test_methyl_data_illumina(raw_data)
            self._test_methyl_data_noob(raw_data)
            self._test_methyl_data_swan(raw_data)
        else:
            for prep in ["raw", "illumina", "swan", "noob"]:
                self.assertIsNotNone(MethylData(raw_data, prep=prep))

    def _test_methyl_data_raw(self, raw_data):
        """Test MethylData with raw preparation."""
        methyl_data = MethylData(raw_data, prep="raw")

        expected_methyl = np.array(
            [
                [18, 1, 14, 0, 9, 17, 6, 14, 9, 20, 13, 17, 8, 13, 6, 19],
                [19, 2, 15, 1, 10, 18, 7, 15, 10, 21, 14, 18, 9, 14, 7, 20],
            ]
        )
        npt.assert_array_equal(methyl_data.methyl, expected_methyl)

        expected_unmethyl = np.array(
            [
                [0, 4, 21, 3, 17, 0, 13, 13, 9, 14, 15, 7, 1, 3, 16, 12],
                [1, 5, 22, 4, 18, 1, 14, 14, 10, 15, 16, 8, 2, 4, 17, 13],
            ]
        )
        npt.assert_array_equal(methyl_data.unmethyl, expected_unmethyl)

    def _test_methyl_data_illumina(self, raw_data):
        """Test MethylData with illumina preparation."""
        methyl_data = MethylData(raw_data, prep="illumina")

        expected_methyl = np.array(
            [
                [1.46428571, 12.90740741, 10.62962963, 19.03571429],
                [2.61702128, 25.15909091, 20.96590909, 18.31914894],
            ]
        )
        npt.assert_almost_equal(
            methyl_data.methyl[:, [1, 5, 7, 13]], expected_methyl
        )

        expected_unmethyl = np.array(
            [
                [3.03703704, 0.0, 9.87037037, 2.27777778],
                [6.98863636, 1.39772727, 19.56818182, 5.59090909],
            ]
        )
        npt.assert_almost_equal(
            methyl_data.unmethyl[:, [1, 5, 7, 13]], expected_unmethyl
        )

    def _test_methyl_data_noob(self, raw_data):
        """Test MethylData with noob preparation."""
        methyl_data = MethylData(raw_data, prep="noob")

        expected_methyl = np.array(
            [
                [17.68105245, 19.39793999, 18.57913213, 19.78231226],
                [17.95140526, 21.91645282, 20.99133582, 20.4679614],
            ]
        )
        npt.assert_almost_equal(
            methyl_data.methyl[:, [1, 5, 7, 13]], expected_methyl
        )

        expected_unmethyl = np.array(
            [
                [16.73134405, 16.36054317, 18.3387984, 16.60061064],
                [18.90364196, 18.36008626, 20.7197986, 18.75593491],
            ]
        )
        npt.assert_almost_equal(
            methyl_data.unmethyl[:, [1, 5, 7, 13]], expected_unmethyl
        )

    def _test_methyl_data_swan(self, raw_data):
        """Test MethylData with swan preparation."""
        methyl_data = MethylData(raw_data, prep="swan", seed=1234)

        expected_methyl = np.array(
            [[5.0, 18.5, 13.5, 13.5], [6.0, 19.5, 14.5, 14.5]]
        )
        npt.assert_almost_equal(
            methyl_data.methyl[:, [1, 5, 7, 13]], expected_methyl
        )

        expected_unmethyl = np.array(
            [[8.42857143, 10.0, 13.0, 5.0], [9.42857143, 11.0, 14.0, 6.0]]
        )
        npt.assert_almost_equal(
            methyl_data.unmethyl[:, [1, 5, 7, 13]], expected_unmethyl
        )


if __name__ == "__main__":
    unittest.main()



"""Writes idat files to disk or to buffer."""

import io

import numpy as np

from mepylome.dtypes.idat import (
    DEFAULT_IDAT_FILE_ID,
    DEFAULT_IDAT_VERSION,
    IdatSectionCode,
)


def write_byte(outfile, value):
    outfile.write(value.to_bytes(1, byteorder="little", signed=False))


def write_short(outfile, value):
    outfile.write(value.to_bytes(2, byteorder="little", signed=False))


def write_int(outfile, value):
    outfile.write(value.to_bytes(4, byteorder="little", signed=True))


def write_long(outfile, value):
    outfile.write(value.to_bytes(8, byteorder="little", signed=True))


def write_char(outfile, value):
    outfile.write(value.encode("utf-8"))


def write_string(outfile, value):
    encoded_value = value.encode("utf-8")
    num_chars = len(encoded_value)
    num_bytes = num_chars
    while num_chars > 127:
        num_bytes = (num_chars & 0x7F) | 0x80
        num_chars >>= 7
        write_byte(outfile, num_bytes)
    write_byte(outfile, num_chars)
    outfile.write(encoded_value)


def write_array(outfile, array):
    outfile.write(array.tobytes())


class IdatWriter:
    """Writes data to an IDAT file with dummy values for missing data."""

    def __init__(self, file=None, data=None):
        """Initializes and writes data to the IDAT file or buffer."""
        self.file = file
        self.data = data
        self.buffer = None
        if self.file:
            with open(self.file, "wb") as outfile:
                self._write_header(outfile)
                self._write_body(outfile)
        else:
            self.buffer = io.BytesIO()
            self._write_header(self.buffer)
            self._write_body(self.buffer)
            self.buffer.seek(0)  # Rewind buffer to the beginning

    def _write_header(self, outfile):
        # Write IDAT header
        write_char(outfile, DEFAULT_IDAT_FILE_ID)
        write_long(outfile, DEFAULT_IDAT_VERSION)
        num_fields = len(IdatSectionCode)
        # Write number of fields
        write_int(outfile, num_fields)
        self._offsets_offset = outfile.tell()
        for _ in IdatSectionCode:
            write_short(outfile, 0)
            write_long(outfile, 0)

    def _get_default(self, n_snps_read):
        return {
            "n_snps_read": n_snps_read,
            "illumina_ids": np.arange(0, n_snps_read, dtype="<i4"),
            "probe_means": np.arange(10, 10 + n_snps_read, dtype="<u2"),
            "std_dev": np.arange(20, 20 + n_snps_read, dtype="<u2"),
            "n_beads": np.arange(30, 30 + n_snps_read, dtype="<u1"),
            "mid_block": np.arange(40, 40 + n_snps_read, dtype="<i4"),
            "run_info": [
                [
                    "run_time_0",
                    "block_type_0",
                    "block_pars_0",
                    "block_code_0",
                    "code_version_0",
                ],
                [
                    "run_time_1",
                    "block_type_1",
                    "block_pars_1",
                    "block_code_1",
                    "code_version_1",
                ],
            ],
            "red_green": 0,
            "mostly_null": "mostly_null",
            "barcode": "barcode",
            "chip_type": "chip_type",
            "mostly_a": "mostly_a",
            "unknown_1": "unknown_1",
            "unknown_2": "unknown_2",
            "unknown_3": "unknown_3",
            "unknown_4": "unknown_4",
            "unknown_5": "unknown_5",
            "unknown_6": "unknown_6",
            "unknown_7": "unknown_7",
        }

    def _write_body(self, outfile):
        # Ensure data is available or use defaults
        illumina_ids = self.data.get("illumina_ids", None)
        n_snps_read = 10 if illumina_ids is None else len(illumina_ids)
        default = self._get_default(n_snps_read)

        self.data = {
            **self.data,
            **{k: v for k, v in default.items() if k not in self.data},
        }

        def check_type(key, dtype):
            if self.data[key].dtype != dtype:
                expected = np.dtype(dtype)
                msg = f"Invalid type: {key} must be of type {expected}."
                raise TypeError(msg)

        check_type("illumina_ids", "<i4")
        check_type("probe_means", "<u2")
        check_type("std_dev", "<u2")
        check_type("n_beads", "<u1")
        check_type("mid_block", "<i4")

        def get_data(key):
            return self.data[key]

        offsets = {}

        # Write NUM_SNPS_READ
        offsets[IdatSectionCode.NUM_SNPS_READ] = outfile.tell()
        write_int(outfile, n_snps_read)

        # Write ILLUMINA_ID
        offsets[IdatSectionCode.ILLUMINA_ID] = outfile.tell()
        write_array(outfile, get_data("illumina_ids"))

        # Write MEAN
        offsets[IdatSectionCode.MEAN] = outfile.tell()
        write_array(outfile, get_data("probe_means"))

        # Write STD_DEV
        offsets[IdatSectionCode.STD_DEV] = outfile.tell()
        write_array(outfile, get_data("std_dev"))

        # Write NUM_BEADS
        offsets[IdatSectionCode.NUM_BEADS] = outfile.tell()
        write_array(outfile, get_data("n_beads"))

        # Write MID_BLOCK
        offsets[IdatSectionCode.MID_BLOCK] = outfile.tell()
        write_int(outfile, len(get_data("mid_block")))
        write_array(outfile, get_data("mid_block"))

        # Write RUN_INFO
        offsets[IdatSectionCode.RUN_INFO] = outfile.tell()
        run_info = get_data("run_info")
        write_int(outfile, len(run_info))
        for entry in run_info:
            for field in entry:
                write_string(outfile, field)

        # Write RED_GREEN
        offsets[IdatSectionCode.RED_GREEN] = outfile.tell()
        write_int(outfile, get_data("red_green"))

        # Write MOSTLY_NULL
        offsets[IdatSectionCode.MOSTLY_NULL] = outfile.tell()
        write_string(outfile, get_data("mostly_null"))

        # Write BARCODE
        offsets[IdatSectionCode.BARCODE] = outfile.tell()
        write_string(outfile, get_data("barcode"))

        # Write CHIP_TYPE
        offsets[IdatSectionCode.CHIP_TYPE] = outfile.tell()
        write_string(outfile, get_data("chip_type"))

        # Write MOSTLY_A
        offsets[IdatSectionCode.MOSTLY_A] = outfile.tell()
        write_string(outfile, get_data("mostly_a"))

        # Write UNKNOWN_1
        offsets[IdatSectionCode.UNKNOWN_1] = outfile.tell()
        write_string(outfile, get_data("unknown_1"))

        # Write UNKNOWN_2
        offsets[IdatSectionCode.UNKNOWN_2] = outfile.tell()
        write_string(outfile, get_data("unknown_2"))

        # Write UNKNOWN_3
        offsets[IdatSectionCode.UNKNOWN_3] = outfile.tell()
        write_string(outfile, get_data("unknown_3"))

        # Write UNKNOWN_4
        offsets[IdatSectionCode.UNKNOWN_4] = outfile.tell()
        write_string(outfile, get_data("unknown_4"))

        # Write UNKNOWN_5
        offsets[IdatSectionCode.UNKNOWN_5] = outfile.tell()
        write_string(outfile, get_data("unknown_5"))

        # Write UNKNOWN_6
        offsets[IdatSectionCode.UNKNOWN_6] = outfile.tell()
        write_string(outfile, get_data("unknown_6"))

        # Write UNKNOWN_7
        offsets[IdatSectionCode.UNKNOWN_7] = outfile.tell()
        write_string(outfile, get_data("unknown_7"))

        # Write offsets
        outfile.seek(self._offsets_offset)
        for code in IdatSectionCode:
            write_short(outfile, code.value)
            write_long(outfile, offsets[code])


def write_idat(data=None, file=None):
    writer = IdatWriter(file=file, data=data)
    if file is None:
        return writer.buffer



"""Helper functions for unittests."""

import gzip
import uuid

from mepylome.tests.write_idat import IdatWriter
from mepylome.utils.files import (
    ensure_directory_exists,
)
from mepylome.utils.varia import MEPYLOME_TMP_DIR

# Create a temporary test directory
TEST_DIR = MEPYLOME_TMP_DIR / "tests"
ensure_directory_exists(TEST_DIR)


def _write_binary(path, value):
    if path.suffix == ".gz":
        with gzip.open(path, "wb") as gz_file:
            gz_file.write(value)
    else:
        with open(path, "wb") as file:
            file.write(value)


class TempIdatFile:
    """Creates a temporary IDAT file."""

    def __init__(self, data, gzipped=False):
        suffix = "idat.gz" if gzipped else ".idat"
        basename = str(uuid.uuid4())
        self.path = TEST_DIR / (basename + suffix)

        idat_writer = IdatWriter(data=data)
        self.data = idat_writer.data

        # Write the IDAT data to the file
        _write_binary(self.path, idat_writer.buffer.getvalue())

    def __del__(self):
        self.path.unlink()


class TempIdatFilePair:
    """Creates a temporary IDAT file pair."""

    def __init__(self, data_grn, data_red, gzipped=False):
        basename = str(uuid.uuid4())
        suffix = "idat.gz" if gzipped else ".idat"
        self.basepath = TEST_DIR / basename
        self.path_grn = TEST_DIR / (basename + "_Grn" + suffix)
        self.path_red = TEST_DIR / (basename + "_Red" + suffix)

        idat_writer_grn = IdatWriter(data=data_grn)
        self.data_grn = idat_writer_grn.data

        idat_writer_red = IdatWriter(data=data_red)
        self.data_red = idat_writer_red.data

        # Write the IDAT data to the file
        _write_binary(self.path_grn, idat_writer_grn.buffer.getvalue())
        _write_binary(self.path_red, idat_writer_red.buffer.getvalue())

    def __del__(self):
        self.path_grn.unlink()
        self.path_red.unlink()


TEST_MANIFEST_CSV = b"""Illumina, Inc.,,,,,,,,,
[Heading],,,,,,,,,,
Descriptor File Name,BS0010894-AQP_content.bpm,,,,,,,,,
Assay Format,Infinium 2,,,,,,,,,
Date Manufactured,6/11/2008,,,,,,,,,
Loci Count ,485553,,,,,,,,,
[Assay],,,,,,,,,,
IlmnID,Name,AddressA_ID,AlleleA_ProbeSeq,AddressB_ID,AlleleB_ProbeSeq,Infinium_Design_Type,Color_Channel,CHR,MAPINFO,Strand
cg00035864,cg00035864,31729416,AAAACACTAACAATCTTATCCACATAAACCCTTAAATTTATCTCAAATTC,,,II,,Y,8553009,F
cg21776599,cg21776599,12751494,ATCTAAATAAAACCTTTCAAAATCCTCCTAAAATCAACATCCRAATAACC,,,II,,1,1830693,F
cg10651537,cg10651537,22700311,ACTTTTTAAATACCAAAACCCAAAAACTACCACCACTCTTAAACTAAAAC,,,II,,2,177134737,F
cg03653856,cg03653856,64649502,ACTAAATTAAAAATTTCTAAAATAACACAAACCCCTCCACTACACTCACA,71783336,ACTAAATTAAAAATTTCTAAAATAACGCGAACCCCTCCGCTACACTCACG,I,Grn,3,197500859,R
cg03738331,cg03738331,29800410,CTACRAAAACAAAAACRAACTAATTCCCTAACCAACCATTAACAAAATCC,,,II,,4,3076305,R
cg07525751,cg07525751,35667488,TTCCAAAAATTATTAAAACACTACAATCATATCATATATAAATAAAACCA,55682328,TTCCGAAAATTATTAAAACGCTACGATCATATCATATATAAATAAAACCG,I,Red,5,1416958,F
cg04704193,cg04704193,14708316,TCCTAAAATATCATCAATAAAACCCAACAAAAAATAAATAATAAAAATCA,52752332,TCCTAAAATATCATCAATAAAACCCAACGAAAAATAAATAATAAAAATCG,I,Red,6,26272200,F
cg17193551,cg17193551,66721465,CAATTTAACTCCAAAACCCRTACTTTATTTATACTAATACTACTCAAAAC,,,II,,8,130925897,F
cg14368972,cg14368972,69739459,TAACAACAAAACAAAACTAACTTCATAACAAATCAATTTCCCTCTCAACA,22775438,TAACAACGAAACAAAACTAACTTCATAACAAATCAATTTCCCTCTCAACG,I,Red,9,2717146,F
cg07175255,cg07175255,72782355,TCAAAACACTAAATAATTAAAAATATCTCTTCAACAAAAAAACAATACCA,11622354,TCAAAACACTAAATAATTAAAAATATCTCTTCGACGAAAAAACGATACCG,I,Red,10,131218009,F
cg04398180,cg04398180,62793332,ATCTTATAAAACRCTAAACACRACTACTTCCTAATTCAAAATCTACTCTC,,,II,,13,114103452,R
cg25598086,cg25598086,48773376,CTCCTAAAATTTTATAATAAATACTTCACTACRTCTCAATAATTATAAAC,,,II,,14,65006720,F
cg08354699,cg08354699,16765420,CTCCATAACCACAAAAAAAACAACTAAAAAAACTACAACTCCAAAACCCA,32725388,CTCCATAACCACGAAAAAAACAACTAAAAAAACTACGACTCCAAAACCCG,I,Red,15,63796895,
cg09913882,cg09913882,18687439,RAATAAACACCRTTAATACCCAACTTAACCAAAACCTCRTCCTTAAAATC,,,II,,17,73612981,F
cg01017773,cg01017773,30736309,ATATCTAAAACAATAAAAACRCTCRTATTTAAAACCCATTTTAACAACAC,,,II,,19,57988497,F
cg27509521,cg27509521,12654498,ACRTAACAACCAATAAACRCATAAATTAACTAAATACRACCAATAAAAAC,,,II,,19,17420279,R
rs10796216,rs10796216,14622465,TAACTAAAAAACAACAATACTAACTCTACACTAAATACCCACTAACCCTC,41635319,TAACTAAAAAACAACAATACTAACTCTACACTAAATACCCACTAACCCTT,I,Red,,,
rs715359,rs715359,18796328,TTATTAAACTCTCACCACTAACTTTCTACTTCTCTCAAAATCAAAACCTC,48710462,TTATTAAACTCTCACCACTAACTTTCTACTTCTCTCAAAATCAAAACCTT,I,Grn,,,
[Controls],,,,,,,,,,
34648333,STAINING,Blue,Biotin (Bkg),,,,,,,
31698466,EXTENSION,Blue,Extension (G),,,,,,,
26772442,HYBRIDIZATION,Blue,Hyb (Medium),,,,,,,
21771417,HYBRIDIZATION,Black,Hyb (Low),,,,,,,
13643320,TARGET REMOVAL,Green,Target Removal 1,,,,,,,
22795447,BISULFITE CONVERSION I,LimeGreen,BS Conversion I-C2,,,,,,,
65797428,SPECIFICITY I,Gold,GT Mismatch 5 (MM),,,,,,,
17661470,SPECIFICITY II,Red,Specificity 2,,,,,,,
70645401,NON-POLYMORPHIC,Blue,NP (G),,,,,,,
13792480,NEGATIVE,Red,Negative 1,AVG,,,,,,
74737439,NEGATIVE,Maroon,Negative 5,,,,,,,
40758350,NEGATIVE,Tan,Negative 9,,,,,,,
74688409,NEGATIVE,Teal,Negative 16,,,,,,,
30689326,NEGATIVE,BlueViolet,Negative 532,,,,,,,
18733503,NORM_G,Blue,Norm_G1,,,,,,,
70603486,NORM_A,Red,Norm_A2,,,,,,,
63775337,NORM_G,Blue,Norm_G2,,,,,,,
70622474,NORM_A,Red,Norm_A3,,,,,,,
58676498,NORM_G,Blue,Norm_G3,,,,,,,
17630478,NORM_A,Red,Norm_A4,,,,,,,
35689446,NORM_G,Blue,Norm_G4,,,,,,,
15639322,NORM_A,Red,Norm_A5,,,,,,,
61760309,NORM_G,Blue,Norm_G5,,,,,,,
66624473,NORM_A,Red,Norm_A6,,,,,,,"""


class TempManifest:
    """Creates a temporary manifest file."""

    def __init__(self):
        basename = str(uuid.uuid4())
        self.path = TEST_DIR / ("tmp_manifest_" + basename + ".csv")
        _write_binary(self.path, TEST_MANIFEST_CSV)

    def __del__(self):
        self.path.unlink()
